@@ -1,5 +1,5 @@
 #-----------------------------------------------------------------------------
-# Copyright (c) 2012 - 2019, Anaconda, Inc., and Bokeh Contributors.
+# Copyright (c) 2012 - 2020, Anaconda, Inc., and Bokeh Contributors.
 # All rights reserved.
 #
 # The full license is in the file LICENSE.txt, distributed with this software.
@@ -11,17 +11,13 @@
 #-----------------------------------------------------------------------------
 # Boilerplate
 #-----------------------------------------------------------------------------
-from __future__ import absolute_import, division, print_function, unicode_literals
-
-import logging
+import logging # isort:skip
 log = logging.getLogger(__name__)
 
 #-----------------------------------------------------------------------------
 # Imports
 #-----------------------------------------------------------------------------
 
-# Standard library imports
-
 # External imports
 from selenium.common.exceptions import TimeoutException
 from selenium.webdriver.common.action_chains import ActionChains
@@ -41,14 +37,20 @@ __all__ = (
     'ButtonWrapper',
     'copy_table_rows',
     'COUNT',
+    'drag_range_slider',
+    'drag_slider',
     'element_to_finish_resizing',
     'element_to_start_resizing',
     'enter_text_in_cell',
     'enter_text_in_cell_with_click_enter',
     'enter_text_in_element',
     'get_page_element',
+    'get_slider_bar_color',
+    'get_slider_title_text',
+    'get_slider_title_value',
     'get_table_cell',
     'get_table_column_cells',
+    'get_table_header',
     'get_table_row',
     'get_table_selected_rows',
     'hover_element',
@@ -57,6 +59,7 @@ __all__ = (
     'RECORD',
     'RESULTS',
     'SCROLL',
+    'select_element_and_press_key',
     'shift_click',
     'sort_table_column',
     'wait_for_canvas_resize',
@@ -71,8 +74,11 @@ def COUNT(key):
 
 INIT = 'Bokeh._testing.init();'
 
-def RECORD(key, value):
-    return 'Bokeh._testing.record(%r, %s);' % (key, value)
+def RECORD(key, value, *, final=True):
+    if final:
+        return 'Bokeh._testing.record(%r, %s);' % (key, value)
+    else:
+        return 'Bokeh._testing.record0(%r, %s);' % (key, value)
 
 RESULTS = 'return Bokeh._testing.results'
 
@@ -91,6 +97,7 @@ def alt_click(driver, element):
     actions.perform()
 
 class ButtonWrapper(object):
+
     def __init__(self, label, callback):
         self.ref = "button-" + make_id()
         self.obj = Button(label=label, css_classes=[self.ref])
@@ -103,6 +110,7 @@ class ButtonWrapper(object):
 class element_to_start_resizing(object):
     ''' An expectation for checking if an element has started resizing
     '''
+
     def __init__(self, element):
         self.element = element
         self.previous_width = self.element.size['width']
@@ -119,6 +127,7 @@ class element_to_finish_resizing(object):
     ''' An expectation for checking if an element has finished resizing
 
     '''
+
     def __init__(self, element):
         self.element = element
         self.previous_width = self.element.size['width']
@@ -131,18 +140,31 @@ class element_to_finish_resizing(object):
             self.previous_width = current_width
             return False
 
+def select_element_and_press_key(driver, element, key, press_number=1):
+    actions = ActionChains(driver)
+    actions.move_to_element(element)
+    actions.click()
+    for i in range(press_number):
+        actions = ActionChains(driver)
+        actions.send_keys_to_element(element, key)
+        actions.perform()
+
 def hover_element(driver, element):
     hover = ActionChains(driver).move_to_element(element)
     hover.perform()
 
-def enter_text_in_element(driver, element, text, click=1, enter=True):
+def enter_text_in_element(driver, element, text, click=1, enter=True, mod=None):
     actions = ActionChains(driver)
     actions.move_to_element(element)
     if click == 1: actions.click()
     elif click == 2: actions.double_click()
     if enter:
         text += Keys.ENTER
+    if mod:
+        actions.key_down(mod)
     actions.send_keys(text)
+    if mod:
+        actions.key_up(mod)
     actions.perform()
 
 def enter_text_in_cell(driver, cell, text):
@@ -173,7 +195,7 @@ def copy_table_rows(driver, rows):
     actions.key_down(Keys.CONTROL)
     actions.send_keys(Keys.INSERT)
     actions.key_up(Keys.CONTROL)
-    #actions.send_keys(Keys.CONTROL, 'c')
+    # actions.send_keys(Keys.CONTROL, 'c')
     actions.perform()
 
 def paste_values(driver, el=None):
@@ -183,7 +205,7 @@ def paste_values(driver, el=None):
     actions.key_down(Keys.SHIFT)
     actions.send_keys(Keys.INSERT)
     actions.key_up(Keys.SHIFT)
-      #actions.send_keys(Keys.CONTROL, 'v')
+    # actions.send_keys(Keys.CONTROL, 'v')
     actions.perform()
 
 def get_table_column_cells(driver, col):
@@ -211,6 +233,9 @@ def get_table_selected_rows(driver):
 def get_table_cell(driver, row, col):
     return driver.find_element_by_css_selector('.grid-canvas .slick-row:nth-child(%d) .r%d' % (row, col))
 
+def get_table_header(driver, col):
+    return driver.find_element_by_css_selector('.slick-header-columns .slick-header-column:nth-child(%d)' % col)
+
 def get_page_element(driver, element_selector):
     return driver.find_element_by_css_selector(element_selector)
 
@@ -240,6 +265,40 @@ def wait_for_canvas_resize(canvas, test_driver):
         # try and process.
         pass
 
+def drag_slider(driver, css_class, distance, release=True):
+    el = driver.find_element_by_css_selector(css_class)
+    handle = el.find_element_by_css_selector('.bk-noUi-handle')
+    actions = ActionChains(driver)
+    actions.move_to_element(handle)
+    actions.click_and_hold()
+    actions.move_by_offset(distance, 0)
+    if release:
+        actions.release()
+    actions.perform()
+
+def drag_range_slider(driver, css_class, location, distance):
+    el = driver.find_element_by_css_selector(css_class)
+    handle = el.find_element_by_css_selector('.bk-noUi-handle-' + location)
+    actions = ActionChains(driver)
+    actions.move_to_element(handle)
+    actions.click_and_hold()
+    actions.move_by_offset(distance, 0)
+    actions.release()
+    actions.perform()
+
+def get_slider_title_text(driver, css_class):
+    el = driver.find_element_by_css_selector(css_class)
+    return el.find_element_by_css_selector('div.bk-input-group > div.bk-slider-title').text
+
+def get_slider_title_value(driver, css_class):
+    el = driver.find_element_by_css_selector(css_class)
+    return el.find_element_by_css_selector('div.bk-input-group > div > span.bk-slider-value').text
+
+def get_slider_bar_color(driver, css_class):
+    el = driver.find_element_by_css_selector(css_class)
+    bar = el.find_element_by_css_selector('.bk-noUi-connect')
+    return bar.value_of_css_property('background-color')
+
 #-----------------------------------------------------------------------------
 # Dev API
 #-----------------------------------------------------------------------------
