@@ -1,7 +1,6 @@
 #-----------------------------------------------------------------------------
-# Copyright (c) 2012 - 2018, Anaconda, Inc. All rights reserved.
-#
-# Powered by the Bokeh Development Team.
+# Copyright (c) 2012 - 2020, Anaconda, Inc., and Bokeh Contributors.
+# All rights reserved.
 #
 # The full license is in the file LICENSE.txt, distributed with this software.
 #-----------------------------------------------------------------------------
@@ -19,9 +18,7 @@ serializable properties.
 #-----------------------------------------------------------------------------
 # Boilerplate
 #-----------------------------------------------------------------------------
-from __future__ import absolute_import, division, print_function, unicode_literals
-
-import logging
+import logging # isort:skip
 log = logging.getLogger(__name__)
 
 #-----------------------------------------------------------------------------
@@ -30,15 +27,10 @@ log = logging.getLogger(__name__)
 
 # Standard library imports
 import difflib
-import inspect
 from warnings import warn
 
-# External imports
-
 # Bokeh imports
-from ..util.future import with_metaclass
 from ..util.string import nice_join
-
 from .property.descriptor_factory import PropertyDescriptorFactory
 from .property.override import Override
 from .property.wrappers import PropertyValueContainer
@@ -156,7 +148,7 @@ class MetaHasProps(type):
             if "__doc__" in class_dict and class_dict["__doc__"] is not None:
                 class_dict["__doc__"] += _EXAMPLE_TEMPLATE % dict(path=path)
 
-        return super(MetaHasProps, meta_cls).__new__(meta_cls, class_name, bases, class_dict)
+        return super().__new__(meta_cls, class_name, bases, class_dict)
 
     def __init__(cls, class_name, bases, nmspc):
         if class_name == 'HasProps':
@@ -205,7 +197,7 @@ def accumulate_from_superclasses(cls, propname):
     # classes, and the cache must be separate for each class
     if cachename not in cls.__dict__:
         s = set()
-        for c in inspect.getmro(cls):
+        for c in cls.__mro__:
             if issubclass(c, HasProps) and hasattr(c, propname):
                 base = getattr(c, propname)
                 s.update(base)
@@ -228,7 +220,7 @@ def accumulate_dict_from_superclasses(cls, propname):
     # classes, and the cache must be separate for each class
     if cachename not in cls.__dict__:
         d = dict()
-        for c in inspect.getmro(cls):
+        for c in cls.__mro__:
             if issubclass(c, HasProps) and hasattr(c, propname):
                 base = getattr(c, propname)
                 for k,v in base.items():
@@ -237,7 +229,7 @@ def accumulate_dict_from_superclasses(cls, propname):
         setattr(cls, cachename, d)
     return cls.__dict__[cachename]
 
-class HasProps(with_metaclass(MetaHasProps, object)):
+class HasProps(object, metaclass=MetaHasProps):
     ''' Base class for all class types that have Bokeh properties.
 
     '''
@@ -245,7 +237,7 @@ class HasProps(with_metaclass(MetaHasProps, object)):
         '''
 
         '''
-        super(HasProps, self).__init__()
+        super().__init__()
         self._property_values = dict()
         self._unstable_default_values = dict()
         self._unstable_themed_values = dict()
@@ -271,14 +263,14 @@ class HasProps(with_metaclass(MetaHasProps, object)):
         # self.properties() below can be expensive so avoid it
         # if we're just setting a private underscore field
         if name.startswith("_"):
-            super(HasProps, self).__setattr__(name, value)
+            super().__setattr__(name, value)
             return
 
         props = sorted(self.properties())
         descriptor = getattr(self.__class__, name, None)
 
         if name in props or (descriptor is not None and descriptor.fset is not None):
-            super(HasProps, self).__setattr__(name, value)
+            super().__setattr__(name, value)
         else:
             matches, text = difflib.get_close_matches(name.lower(), props), "similar"
 
@@ -600,7 +592,7 @@ class HasProps(with_metaclass(MetaHasProps, object)):
         old_dict = self.themed_values()
 
         # if the same theme is set again, it should reuse the same dict
-        if old_dict is property_values:
+        if old_dict is property_values:  # lgtm [py/comparison-using-is]
             return
 
         removed = set()
@@ -657,12 +649,13 @@ _ABSTRACT_ADMONITION = '''
 
 '''
 
+# The "../../" is needed for bokeh-plot to construct the correct path to examples
 _EXAMPLE_TEMPLATE = '''
 
     Example
     -------
 
-    .. bokeh-plot:: ../%(path)s
+    .. bokeh-plot:: ../../%(path)s
         :source-position: below
 
 '''
