@@ -1,5 +1,5 @@
 #-----------------------------------------------------------------------------
-# Copyright (c) 2012 - 2019, Anaconda, Inc., and Bokeh Contributors.
+# Copyright (c) 2012 - 2020, Anaconda, Inc., and Bokeh Contributors.
 # All rights reserved.
 #
 # The full license is in the file LICENSE.txt, distributed with this software.
@@ -93,7 +93,9 @@ Basic Properties
 .. autoclass:: JSON
 .. autoclass:: MarkerType
 .. autoclass:: MinMaxBounds
+.. autoclass:: NonNegativeInt
 .. autoclass:: Percent
+.. autoclass:: PositiveInt
 .. autoclass:: RGB
 .. autoclass:: Regex
 .. autoclass:: Size
@@ -123,6 +125,7 @@ DataSpec Properties
 .. autoclass:: FontSizeSpec
 .. autoclass:: MarkerSpec
 .. autoclass:: NumberSpec
+.. autoclass:: PropertyUnitsSpec
 .. autoclass:: ScreenDistanceSpec
 .. autoclass:: StringSpec
 .. autoclass:: UnitsSpec
@@ -140,6 +143,12 @@ Special Properties
 .. autoclass:: Include
 .. autoclass:: Override
 
+Validation-only Properties
+--------------------------
+
+.. autoclass:: PandasDataFrame
+.. autoclass:: PandasGroupBy
+
 Validation Control
 ------------------
 
@@ -158,21 +167,13 @@ to control when type validation occurs.
 #-----------------------------------------------------------------------------
 # Boilerplate
 #-----------------------------------------------------------------------------
-from __future__ import absolute_import, division, print_function, unicode_literals
-
-import logging
+import logging # isort:skip
 log = logging.getLogger(__name__)
 
 #-----------------------------------------------------------------------------
 # Imports
 #-----------------------------------------------------------------------------
 
-# Standard library imports
-
-# External imports
-
-# Bokeh imports
-
 #-----------------------------------------------------------------------------
 # Globals and constants
 #-----------------------------------------------------------------------------
@@ -215,9 +216,14 @@ __all__ = (
     'MarkerSpec',
     'MarkerType',
     'MinMaxBounds',
+    'NonNegativeInt',
     'NumberSpec',
     'Override',
+    'PandasDataFrame',
+    'PandasGroupBy',
     'Percent',
+    'PositiveInt',
+    'PropertyUnitsSpec',
     'RGB',
     'Regex',
     'RelativeDelta',
@@ -269,6 +275,7 @@ from .property.dataspec import FontSizeSpec; FontSizeSpec
 from .property.dataspec import HatchPatternSpec; HatchPatternSpec
 from .property.dataspec import MarkerSpec; MarkerSpec
 from .property.dataspec import NumberSpec; NumberSpec
+from .property.dataspec import PropertyUnitsSpec; PropertyUnitsSpec
 from .property.dataspec import ScreenDistanceSpec; ScreenDistanceSpec
 from .property.dataspec import StringSpec; StringSpec
 from .property.dataspec import UnitsSpec; UnitsSpec
@@ -293,10 +300,14 @@ from .property.numeric import Byte; Byte
 from .property.numeric import Interval; Interval
 from .property.numeric import NonNegativeInt; NonNegativeInt
 from .property.numeric import Percent; Percent
+from .property.numeric import PositiveInt; PositiveInt
 from .property.numeric import Size; Size
 
 from .property.override import Override ; Override
 
+from .property.pandas import PandasDataFrame ; PandasDataFrame
+from .property.pandas import PandasGroupBy ; PandasGroupBy
+
 from .property.primitive import Bool; Bool
 from .property.primitive import Complex; Complex
 from .property.primitive import Int; Int
