@@ -1,7 +1,6 @@
 #-----------------------------------------------------------------------------
-# Copyright (c) 2012 - 2018, Anaconda, Inc. All rights reserved.
-#
-# Powered by the Bokeh Development Team.
+# Copyright (c) 2012 - 2020, Anaconda, Inc., and Bokeh Contributors.
+# All rights reserved.
 #
 # The full license is in the file LICENSE.txt, distributed with this software.
 #-----------------------------------------------------------------------------
@@ -12,20 +11,13 @@
 #-----------------------------------------------------------------------------
 # Boilerplate
 #-----------------------------------------------------------------------------
-from __future__ import absolute_import, division, print_function, unicode_literals
-
-import logging
+import logging # isort:skip
 log = logging.getLogger(__name__)
 
 #-----------------------------------------------------------------------------
 # Imports
 #-----------------------------------------------------------------------------
 
-# Standard library imports
-
-# External imports
-from six import string_types
-
 # Bokeh imports
 from ... import colors
 from ...util.serialization import convert_datetime_type, convert_timedelta_type
@@ -38,7 +30,7 @@ from .either import Either
 from .enum import Enum
 from .instance import Instance
 from .primitive import Float, String
-from .visual import FontSize, MarkerType
+from .visual import FontSize, HatchPatternType, MarkerType
 
 #-----------------------------------------------------------------------------
 # Globals and constants
@@ -53,6 +45,7 @@ __all__ = (
     'expr',
     'field',
     'FontSizeSpec',
+    'HatchPatternSpec',
     'MarkerSpec',
     'NumberSpec',
     'ScreenDistanceSpec',
@@ -140,7 +133,7 @@ class DataSpec(Either):
         glyph.x = { 'value': 10, 'transform': Jitter(width=0.4) }
 
     Note that ``DataSpec`` is not normally useful on its own. Typically,
-    a model will define properties using one of the sublclasses such
+    a model will define properties using one of the subclasses such
     as :class:`~bokeh.core.properties.NumberSpec` or
     :class:`~bokeh.core.properties.ColorSpec`. For example, a Bokeh
     model with ``x``, ``y`` and ``color`` properties that can handle
@@ -158,7 +151,7 @@ class DataSpec(Either):
 
     '''
     def __init__(self, key_type, value_type, default, help=None):
-        super(DataSpec, self).__init__(
+        super().__init__(
             String,
             Dict(
                 key_type,
@@ -205,7 +198,7 @@ class DataSpec(Either):
             pass
 
         # Check for data source field name
-        if isinstance(val, string_types):
+        if isinstance(val, str):
             return dict(field=val)
 
         # Must be dict, return a new dict
@@ -236,7 +229,7 @@ class NumberSpec(DataSpec):
 
     '''
     def __init__(self, default=None, help=None, key_type=_ExprFieldValueTransform, accept_datetime=True, accept_timedelta=True):
-        super(NumberSpec, self).__init__(key_type, Float, default=default, help=help)
+        super().__init__(key_type, Float, default=default, help=help)
         if accept_timedelta:
             self.accepts(TimeDelta, convert_timedelta_type)
         if accept_datetime:
@@ -259,14 +252,14 @@ class StringSpec(DataSpec):
 
     '''
     def __init__(self, default, help=None, key_type=_ExprFieldValueTransform):
-        super(StringSpec, self).__init__(key_type, List(String), default=default, help=help)
+        super().__init__(key_type, List(String), default=default, help=help)
 
     def prepare_value(self, cls, name, value):
         if isinstance(value, list):
             if len(value) != 1:
                 raise TypeError("StringSpec convenience list values must have length 1")
             value = dict(value=value[0])
-        return super(StringSpec, self).prepare_value(cls, name, value)
+        return super().prepare_value(cls, name, value)
 
 class FontSizeSpec(DataSpec):
     ''' A |DataSpec| property that accepts font-size fixed values.
@@ -290,17 +283,37 @@ class FontSizeSpec(DataSpec):
     '''
 
     def __init__(self, default, help=None, key_type=_ExprFieldValueTransform):
-        super(FontSizeSpec, self).__init__(key_type, FontSize, default=default, help=help)
+        super().__init__(key_type, FontSize, default=default, help=help)
 
     def validate(self, value, detail=True):
         # We want to preserve existing semantics and be a little more restrictive. This
         # validations makes m.font_size = "" or m.font_size = "6" an error
-        super(FontSizeSpec, self).validate(value, detail)
-        if isinstance(value, string_types):
+        super().validate(value, detail)
+        if isinstance(value, str):
             if len(value) == 0 or value[0].isdigit() and FontSize._font_size_re.match(value) is None:
                 msg = "" if not detail else "%r is not a valid font size value" % value
                 raise ValueError(msg)
 
+class HatchPatternSpec(DataSpec):
+    ''' A |DataSpec| property that accepts hatch pattern types as fixed values.
+
+    The ``HatchPatternSpec`` property attempts to first interpret string values
+    as hatch pattern types. Otherwise string values are interpreted as field
+    names. For example:
+
+    .. code-block:: python
+
+        m.font_size = "."    # value
+
+        m.font_size = "ring" # value
+
+        m.font_size = "foo"  # field
+
+    '''
+
+    def __init__(self, default, help=None, key_type=_ExprFieldValueTransform):
+        super().__init__(key_type, HatchPatternType, default=default, help=help)
+
 class MarkerSpec(DataSpec):
     ''' A |DataSpec| property that accepts marker types as fixed values.
 
@@ -319,27 +332,67 @@ class MarkerSpec(DataSpec):
     '''
 
     def __init__(self, default, help=None, key_type=_ExprFieldValueTransform):
-        super(MarkerSpec, self).__init__(key_type, MarkerType, default=default, help=help)
+        super().__init__(key_type, MarkerType, default=default, help=help)
 
 
 class UnitsSpec(NumberSpec):
     ''' A |DataSpec| property that accepts numeric fixed values, and also
-    provides an associated units property to store units information.
+    serializes associated units values.
 
     '''
     def __init__(self, default, units_type, units_default, help=None):
-        super(UnitsSpec, self).__init__(default=default, help=help, key_type=_ExprFieldValueTransformUnits)
+        super().__init__(default=default, help=help, key_type=_ExprFieldValueTransformUnits)
         self._units_type = self._validate_type_param(units_type)
-        # this is a hack because we already constructed units_type
+
+        # TODO (bev) units_type was already constructed, so this really should not be needed
         self._units_type.validate(units_default)
         self._units_type._default = units_default
-        # this is sort of a hack because we don't have a
-        # serialized= kwarg on every Property subtype
         self._units_type._serialized = False
 
     def __str__(self):
         return "%s(units_default=%r)" % (self.__class__.__name__, self._units_type._default)
 
+    def get_units(self, obj, name):
+        raise NotImplementedError()
+
+    def make_descriptors(self, base_name):
+        ''' Return a list of ``PropertyDescriptor`` instances to install on a
+        class, in order to delegate attribute access to this property.
+
+        Unlike simpler property types, ``UnitsSpec`` returns multiple
+        descriptors to install. In particular, descriptors for the base
+        property as well as the associated units property are returned.
+
+        Args:
+            name (str) : the name of the property these descriptors are for
+
+        Returns:
+            list[PropertyDescriptor]
+
+        The descriptors returned are collected by the ``MetaHasProps``
+        metaclass and added to ``HasProps`` subclasses during class creation.
+        '''
+        units_props = self._units_type.make_descriptors("unused")
+        return [ UnitsSpecPropertyDescriptor(base_name, self, units_props[0]) ]
+
+    def to_serializable(self, obj, name, val):
+        d = super().to_serializable(obj, name, val)
+        if d is not None and 'units' not in d:
+            # d is a PropertyValueDict at this point, we need to convert it to
+            # a plain dict if we are going to modify its value, otherwise a
+            # notify_change that should not happen will be triggered
+            d = dict(d)
+            d["units"] = self.get_units(obj, name)
+        return d
+
+class PropertyUnitsSpec(UnitsSpec):
+    ''' A |DataSpec| property that accepts numeric fixed values, and also
+    provides an associated units property to store units information.
+
+    '''
+    def get_units(self, obj, name):
+        return getattr(obj, name+"_units")
+
     def make_descriptors(self, base_name):
         ''' Return a list of ``PropertyDescriptor`` instances to install on a
         class, in order to delegate attribute access to this property.
@@ -361,17 +414,7 @@ class UnitsSpec(NumberSpec):
         units_props = self._units_type.make_descriptors(units_name)
         return units_props + [ UnitsSpecPropertyDescriptor(base_name, self, units_props[0]) ]
 
-    def to_serializable(self, obj, name, val):
-        d = super(UnitsSpec, self).to_serializable(obj, name, val)
-        if d is not None and 'units' not in d:
-            # d is a PropertyValueDict at this point, we need to convert it to
-            # a plain dict if we are going to modify its value, otherwise a
-            # notify_change that should not happen will be triggered
-            d = dict(d)
-            d["units"] = getattr(obj, name+"_units")
-        return d
-
-class AngleSpec(UnitsSpec):
+class AngleSpec(PropertyUnitsSpec):
     ''' A |DataSpec| property that accepts numeric fixed values, and also
     provides an associated units property to store angle units.
 
@@ -379,9 +422,9 @@ class AngleSpec(UnitsSpec):
 
     '''
     def __init__(self, default=None, units_default="rad", help=None):
-        super(AngleSpec, self).__init__(default=default, units_type=Enum(enums.AngleUnits), units_default=units_default, help=help)
+        super().__init__(default=default, units_type=Enum(enums.AngleUnits), units_default=units_default, help=help)
 
-class DistanceSpec(UnitsSpec):
+class DistanceSpec(PropertyUnitsSpec):
     ''' A |DataSpec| property that accepts numeric fixed values or strings
     that refer to columns in a :class:`~bokeh.models.sources.ColumnDataSource`,
     and also provides an associated units property to store units information.
@@ -389,7 +432,7 @@ class DistanceSpec(UnitsSpec):
 
     '''
     def __init__(self, default=None, units_default="data", help=None):
-        super(DistanceSpec, self).__init__(default=default, units_type=Enum(enums.SpatialUnits), units_default=units_default, help=help)
+        super().__init__(default=default, units_type=Enum(enums.SpatialUnits), units_default=units_default, help=help)
 
     def prepare_value(self, cls, name, value):
         try:
@@ -397,109 +440,39 @@ class DistanceSpec(UnitsSpec):
                 raise ValueError("Distances must be positive or None!")
         except TypeError:
             pass
-        return super(DistanceSpec, self).prepare_value(cls, name, value)
+        return super().prepare_value(cls, name, value)
 
-class ScreenDistanceSpec(UnitsSpec):
-    ''' A |DataSpec| property that accepts numeric fixed values for screen
+class _FixedUnitsDistanceSpec(UnitsSpec):
+
+    def __init__(self, default=None, help=None):
+        super().__init__(default=default, units_type=Enum(enums.enumeration(self._units)), units_default=self._units, help=help)
+
+    def get_units(self, _obj, _name):
+        return self._units
+
+    def prepare_value(self, cls, name, value):
+        try:
+            if value is not None and value < 0:
+                raise ValueError("Distances must be positive or None!")
+        except TypeError:
+            pass
+        return super().prepare_value(cls, name, value)
+
+class ScreenDistanceSpec(_FixedUnitsDistanceSpec):
+    ''' A |DataSpec| property that accepts numeric fixed values for screen-space
     distances, and also provides an associated units property that reports
     ``"screen"`` as the units.
 
-    .. note::
-        Units are always ``"screen"``.
-
     '''
+    _units = "screen"
 
-    def __init__(self, default=None, help=None):
-        super(ScreenDistanceSpec, self).__init__(default=default, units_type=Enum(enums.enumeration("screen")), units_default="screen", help=help)
-
-    def prepare_value(self, cls, name, value):
-        try:
-            if value is not None and value < 0:
-                raise ValueError("Distances must be positive or None!")
-        except TypeError:
-            pass
-        return super(ScreenDistanceSpec, self).prepare_value(cls, name, value)
-
-    def make_descriptors(self, base_name):
-        ''' Return a list of ``PropertyDescriptor`` instances to install on a
-        class, in order to delegate attribute access to this property.
-
-        Unlike simpler property types, ``UnitsSpec`` returns multiple
-        descriptors to install. In particular, descriptors for the base
-        property as well as the associated units property are returned.
-
-        Args:
-            name (str) : the name of the property these descriptors are for
-
-        Returns:
-            list[PropertyDescriptor]
-
-        The descriptors returned are collected by the ``MetaHasProps``
-        metaclass and added to ``HasProps`` subclasses during class creation.
-        '''
-        units_props = self._units_type.make_descriptors("unused")
-        return [ UnitsSpecPropertyDescriptor(base_name, self, units_props[0]) ]
-
-    def to_serializable(self, obj, name, val):
-        d = super(UnitsSpec, self).to_serializable(obj, name, val)
-        if d is not None and 'units' not in d:
-            # d is a PropertyValueDict at this point, we need to convert it to
-            # a plain dict if we are going to modify its value, otherwise a
-            # notify_change that should not happen will be triggered
-            d = dict(d)
-            d["units"] = "screen"
-        return d
-
-
-class DataDistanceSpec(UnitsSpec):
+class DataDistanceSpec(_FixedUnitsDistanceSpec):
     ''' A |DataSpec| property that accepts numeric fixed values for data-space
     distances, and also provides an associated units property that reports
     ``"data"`` as the units.
 
-    .. note::
-        Units are always ``"data"``.
-
     '''
-    def __init__(self, default=None, help=None):
-        super(DataDistanceSpec, self).__init__(default=default, units_type=Enum(enums.enumeration("data")), units_default="data", help=help)
-
-    def prepare_value(self, cls, name, value):
-        try:
-            if value is not None and value < 0:
-                raise ValueError("Distances must be positive or None!")
-        except TypeError:
-            pass
-        return super(DataDistanceSpec, self).prepare_value(cls, name, value)
-
-    def make_descriptors(self, base_name):
-        ''' Return a list of ``PropertyDescriptor`` instances to install on a
-        class, in order to delegate attribute access to this property.
-
-        Unlike simpler property types, ``UnitsSpec`` returns multiple
-        descriptors to install. In particular, descriptors for the base
-        property as well as the associated units property are returned.
-
-        Args:
-            name (str) : the name of the property these descriptors are for
-
-        Returns:
-            list[PropertyDescriptor]
-
-        The descriptors returned are collected by the ``MetaHasProps``
-        metaclass and added to ``HasProps`` subclasses during class creation.
-        '''
-        units_props = self._units_type.make_descriptors("unused")
-        return [ UnitsSpecPropertyDescriptor(base_name, self, units_props[0]) ]
-
-    def to_serializable(self, obj, name, val):
-        d = super(UnitsSpec, self).to_serializable(obj, name, val)
-        if d is not None and 'units' not in d:
-            # d is a PropertyValueDict at this point, we need to convert it to
-            # a plain dict if we are going to modify its value, otherwise a
-            # notify_change that should not happen will be triggered
-            d = dict(d)
-            d["units"] = "data"
-        return d
+    _units = "data"
 
 class ColorSpec(DataSpec):
     ''' A |DataSpec| property that accepts |Color| fixed values.
@@ -527,7 +500,7 @@ class ColorSpec(DataSpec):
 
     '''
     def __init__(self, default, help=None, key_type=_ExprFieldValueTransform):
-        super(ColorSpec, self).__init__(key_type, Color, default=default, help=help)
+        super().__init__(key_type, Color, default=default, help=help)
 
     @classmethod
     def isconst(cls, val):
@@ -542,7 +515,7 @@ class ColorSpec(DataSpec):
             True, if the value is a string color literal
 
         '''
-        return isinstance(val, string_types) and \
+        return isinstance(val, str) and \
                ((len(val) == 7 and val[0] == "#") or val in enums.NamedColor)
 
     def to_serializable(self, obj, name, val):
@@ -562,7 +535,7 @@ class ColorSpec(DataSpec):
             return val.to_css()
 
         # Check for data source field name or rgb(a) string
-        if isinstance(val, string_types):
+        if isinstance(val, str):
             if val.startswith(("rgb(", "rgba(")):
                 return val
 
@@ -580,10 +553,9 @@ class ColorSpec(DataSpec):
         # at this point, since Color is very strict about only accepting
         # tuples of (integer) bytes. This conditions tuple values to only
         # have integer RGB components
-        if isinstance(value, tuple):
-            # TODO (bev) verify that all original floats are integer values?
+        if isinstance(value, tuple) and len(value) in (3, 4) and all(isinstance(v, (float, int)) for v in value):
             value = tuple(int(v) if i < 3 else v for i, v in enumerate(value))
-        return super(ColorSpec, self).prepare_value(cls, name, value)
+        return super().prepare_value(cls, name, value)
 
 # DataSpec helpers ------------------------------------------------------------
 
