@@ -1,5 +1,5 @@
 #-----------------------------------------------------------------------------
-# Copyright (c) 2012 - 2019, Anaconda, Inc., and Bokeh Contributors.
+# Copyright (c) 2012 - 2020, Anaconda, Inc., and Bokeh Contributors.
 # All rights reserved.
 #
 # The full license is in the file LICENSE.txt, distributed with this software.
@@ -11,9 +11,7 @@
 #-----------------------------------------------------------------------------
 # Boilerplate
 #-----------------------------------------------------------------------------
-from __future__ import absolute_import, division, print_function, unicode_literals
-
-import logging
+import logging # isort:skip
 log = logging.getLogger(__name__)
 
 #-----------------------------------------------------------------------------
@@ -21,21 +19,23 @@ log = logging.getLogger(__name__)
 #-----------------------------------------------------------------------------
 
 # Standard library imports
+from os.path import dirname, exists, join
 from warnings import warn
 
-# External imports
-
 # Bokeh imports
+from ..core.templates import CSS_RESOURCES, JS_RESOURCES
 from ..document.document import Document
-from ..resources import BaseResources
+from ..model import Model
+from ..resources import BaseResources, Resources
+from ..settings import settings
 from ..util.compiler import bundle_models
-from .wrappers import wrap_in_script_tag
 
 #-----------------------------------------------------------------------------
 # Globals and constants
 #-----------------------------------------------------------------------------
 
 __all__ = (
+    'Bundle',
     'bundle_for_objs_and_resources',
 )
 
@@ -47,6 +47,75 @@ __all__ = (
 # Dev API
 #-----------------------------------------------------------------------------
 
+class ScriptRef(object):
+
+    def __init__(self, url, type="text/javascript"):
+        self.url = url
+        self.type = type
+
+class Script(object):
+
+    def __init__(self, content, type="text/javascript"):
+        self.content = content
+        self.type = type
+
+class StyleRef(object):
+
+    def __init__(self, url):
+        self.url = url
+
+class Style(object):
+
+    def __init__(self, content):
+        self.content = content
+
+class Bundle(object):
+
+    @classmethod
+    def of(cls, js_files, js_raw, css_files, css_raw, hashes):
+        return cls(js_files=js_files, js_raw=js_raw, css_files=css_files, css_raw=css_raw, hashes=hashes)
+
+    def __init__(self, **kwargs):
+        self.js_files = kwargs.get("js_files", [])
+        self.js_raw = kwargs.get("js_raw", [])
+        self.css_files = kwargs.get("css_files", [])
+        self.css_raw = kwargs.get("css_raw", [])
+        self.hashes = kwargs.get("hashes", None)
+
+    def __iter__(self):
+        yield self._render_js()
+        yield self._render_css()
+
+    def _render_js(self):
+        return JS_RESOURCES.render(js_files=self.js_files, js_raw=self.js_raw, hashes=self.hashes)
+
+    def _render_css(self):
+        return CSS_RESOURCES.render(css_files=self.css_files, css_raw=self.css_raw)
+
+    def scripts(self, tag=True):
+        if tag:
+            return JS_RESOURCES.render(js_raw=self.js_raw, js_files=[])
+        else:
+            return "\n".join(self.js_raw)
+
+    @property
+    def js_urls(self):
+        return self.js_files
+
+    @property
+    def css_urls(self):
+        return self.css_files
+
+    def add(self, artifact):
+        if isinstance(artifact, ScriptRef):
+            self.js_files.append(artifact.url)
+        elif isinstance(artifact, Script):
+            self.js_raw.append(artifact.content)
+        elif isinstance(artifact, StyleRef):
+            self.css_files.append(artifact.url)
+        elif isinstance(artifact, Style):
+            self.css_raw.append(artifact.content)
+
 def bundle_for_objs_and_resources(objs, resources):
     ''' Generate rendered CSS and JS resources suitable for the given
     collection of Bokeh objects
@@ -57,10 +126,15 @@ def bundle_for_objs_and_resources(objs, resources):
         resources (BaseResources or tuple[BaseResources])
 
     Returns:
-        tuple
+        Bundle
 
     '''
-    if isinstance(resources, BaseResources):
+    # Any env vars will overide a local default passed in
+    resources = settings.resources(default=resources)
+    if isinstance(resources, str):
+        resources = Resources(mode=resources)
+
+    if resources is None or isinstance(resources, BaseResources):
         js_resources = css_resources = resources
     elif isinstance(resources, tuple) and len(resources) == 2 and all(r is None or isinstance(r, BaseResources) for r in resources):
         js_resources, css_resources = resources
@@ -80,6 +154,11 @@ def bundle_for_objs_and_resources(objs, resources):
     use_tables  = _use_tables(objs)  if objs else True
     use_gl      = _use_gl(objs)      if objs else True
 
+    js_files = []
+    js_raw = []
+    css_files = []
+    css_raw = []
+
     if js_resources:
         js_resources = deepcopy(js_resources)
         if not use_widgets and "bokeh-widgets" in js_resources.js_components:
@@ -88,37 +167,70 @@ def bundle_for_objs_and_resources(objs, resources):
             js_resources.js_components.remove("bokeh-tables")
         if not use_gl and "bokeh-gl" in js_resources.js_components:
             js_resources.js_components.remove("bokeh-gl")
-        bokeh_js = js_resources.render_js()
-    else:
-        bokeh_js = None
 
-    models = [ obj.__class__ for obj in _all_objs(objs) ] if objs else None
-    custom_bundle = bundle_models(models)
-
-    if custom_bundle is not None:
-        custom_bundle = wrap_in_script_tag(custom_bundle)
-
-        if bokeh_js is not None:
-            bokeh_js += "\n" + custom_bundle
-        else:
-            bokeh_js = custom_bundle
+        js_files.extend(js_resources.js_files)
+        js_raw.extend(js_resources.js_raw)
 
     if css_resources:
         css_resources = deepcopy(css_resources)
-        if not use_widgets and "bokeh-widgets" in css_resources.css_components:
-            css_resources.css_components.remove("bokeh-widgets")
-        if not use_tables and "bokeh-tables" in css_resources.css_components:
-            css_resources.css_components.remove("bokeh-tables")
-        bokeh_css = css_resources.render_css()
-    else:
-        bokeh_css = None
+        css_files.extend(css_resources.css_files)
+        css_raw.extend(css_resources.css_raw)
 
-    return bokeh_js, bokeh_css
+    js_raw.extend(_bundle_extensions(objs, resources))
+
+    models = [ obj.__class__ for obj in _all_objs(objs) ] if objs else None
+    ext = bundle_models(models)
+    if ext is not None:
+        js_raw.append(ext)
+
+    return Bundle.of(js_files, js_raw, css_files, css_raw, js_resources.hashes if js_resources else None)
 
 #-----------------------------------------------------------------------------
 # Private API
 #-----------------------------------------------------------------------------
 
+def _query_extensions(objs, query):
+    names = set()
+
+    for obj in _all_objs(objs):
+        if hasattr(obj, "__implementation__"):
+            continue
+        name = obj.__view_module__.split(".")[0]
+        if name == "bokeh":
+            continue
+        if name in names:
+            continue
+        names.add(name)
+
+        for model in Model.model_class_reverse_map.values():
+            if model.__module__.startswith(name):
+                if query(model):
+                    return True
+
+    return False
+
+def _bundle_extensions(objs, resources):
+    names = set()
+    extensions = []
+
+    for obj in _all_objs(objs) if objs is not None else Model.model_class_reverse_map.values():
+        if hasattr(obj, "__implementation__"):
+            continue
+        name = obj.__view_module__.split(".")[0]
+        if name == "bokeh":
+            continue
+        if name in names:
+            continue
+        names.add(name)
+        module = __import__(name)
+        ext = ".min.js" if resources is None or resources.minified else ".js"
+        artifact = join(dirname(module.__file__), "dist", name + ext)
+        if exists(artifact):
+            bundle = BaseResources._inline(artifact)
+            extensions.append(bundle)
+
+    return extensions
+
 def _all_objs(objs):
     all_objs = set()
 
@@ -150,8 +262,7 @@ def _any(objs, query):
         else:
             if any(query(ref) for ref in obj.references()):
                 return True
-    else:
-        return False
+    return False
 
 def _use_gl(objs):
     ''' Whether a collection of Bokeh objects contains a plot requesting WebGL
@@ -177,7 +288,7 @@ def _use_tables(objs):
 
     '''
     from ..models.widgets import TableWidget
-    return _any(objs, lambda obj: isinstance(obj, TableWidget))
+    return _any(objs, lambda obj: isinstance(obj, TableWidget)) or _ext_use_tables(objs)
 
 def _use_widgets(objs):
     ''' Whether a collection of Bokeh objects contains a any Widget
@@ -190,7 +301,15 @@ def _use_widgets(objs):
 
     '''
     from ..models.widgets import Widget
-    return _any(objs, lambda obj: isinstance(obj, Widget))
+    return _any(objs, lambda obj: isinstance(obj, Widget)) or _ext_use_widgets(objs)
+
+def _ext_use_tables(objs):
+    from ..models.widgets import TableWidget
+    return _query_extensions(objs, lambda cls: issubclass(cls, TableWidget))
+
+def _ext_use_widgets(objs):
+    from ..models.widgets import Widget
+    return _query_extensions(objs, lambda cls: issubclass(cls, Widget))
 
 #-----------------------------------------------------------------------------
 # Code
