@@ -1,26 +1,97 @@
+#-----------------------------------------------------------------------------
+# Copyright (c) 2012 - 2020, Anaconda, Inc., and Bokeh Contributors.
+# All rights reserved.
+#
+# The full license is in the file LICENSE.txt, distributed with this software.
+#-----------------------------------------------------------------------------
 ''' Provide a base class for all objects (called Bokeh Models) that can go in
 a Bokeh |Document|.
 
 '''
-from __future__ import absolute_import, print_function
+#-----------------------------------------------------------------------------
+# Boilerplate
+#-----------------------------------------------------------------------------
+import logging # isort:skip
+log = logging.getLogger(__name__)
 
-import logging
-logger = logging.getLogger(__name__)
+#-----------------------------------------------------------------------------
+# Imports
+#-----------------------------------------------------------------------------
 
+# Standard library imports
+from inspect import isclass
 from json import loads
 from operator import itemgetter
 
-from six import iteritems
-
+# Bokeh imports
+from .core.has_props import HasProps, abstract
 from .core.json_encoder import serialize_json
 from .core.properties import Any, Dict, Instance, List, String
-from .core.has_props import HasProps, MetaHasProps
-from .core.query import find
-from .themes import default as default_theme
-from .util.callback_manager import PropertyCallbackManager, EventCallbackManager
-from .util.future import with_metaclass
-from .util.serialization import make_id
 from .events import Event
+from .themes import default as default_theme
+from .util.callback_manager import EventCallbackManager, PropertyCallbackManager
+from .util.serialization import make_id
+
+#-----------------------------------------------------------------------------
+# Globals and constants
+#-----------------------------------------------------------------------------
+
+__all__ = (
+    'collect_models',
+    'get_class',
+    'Model',
+)
+
+#-----------------------------------------------------------------------------
+# General API
+#-----------------------------------------------------------------------------
+
+#-----------------------------------------------------------------------------
+# Dev API
+#-----------------------------------------------------------------------------
+
+def collect_filtered_models(discard, *input_values):
+    ''' Collect a duplicate-free list of all other Bokeh models referred to by
+    this model, or by any of its references, etc, unless filtered-out by the
+    provided callable.
+
+    Iterate over ``input_values`` and descend through their structure
+    collecting all nested ``Models`` on the go.
+
+    Args:
+        *discard (Callable[[Model], bool])
+            a callable which accepts a *Model* instance as its single argument
+            and returns a boolean stating whether to discard the instance. The
+            latter means that the instance will not be added to collected
+            models nor will its references be explored.
+
+        *input_values (Model)
+            Bokeh models to collect other models from
+
+    Returns:
+        None
+
+    '''
+
+    ids = set([])
+    collected = []
+    queued = []
+
+    def queue_one(obj):
+        if obj.id not in ids and not (callable(discard) and discard(obj)):
+            queued.append(obj)
+
+    for value in input_values:
+        _visit_value_and_its_immediate_references(value, queue_one)
+
+    while queued:
+        obj = queued.pop(0)
+        if obj.id not in ids:
+            ids.add(obj.id)
+            collected.append(obj)
+            _visit_immediate_value_references(obj, queue_one)
+
+    return collected
 
 def collect_models(*input_values):
     ''' Collect a duplicate-free list of all other Bokeh models referred to by
@@ -38,26 +109,7 @@ def collect_models(*input_values):
         list[Model] : all models reachable from this one.
 
     '''
-
-    ids = set([])
-    collected = []
-    queued = []
-
-    def queue_one(obj):
-        if obj._id not in ids:
-            queued.append(obj)
-
-    for value in input_values:
-        _visit_value_and_its_immediate_references(value, queue_one)
-
-    while queued:
-        obj = queued.pop(0)
-        if obj._id not in ids:
-            ids.add(obj._id)
-            collected.append(obj)
-            _visit_immediate_value_references(obj, queue_one)
-
-    return collected
+    return collect_filtered_models(None, *input_values)
 
 def get_class(view_model_name):
     ''' Look up a Bokeh model class, given its view model name.
@@ -82,113 +134,21 @@ def get_class(view_model_name):
 
     '''
 
-    # in order to look up from the model catalog that MetaModel maintains, it
+    # in order to look up from the model catalog that Model maintains, it
     # has to be creates first. These imports ensure that all built-in Bokeh
     # models are represented in the catalog.
     from . import models; models
     from .plotting import Figure; Figure
 
-    d = MetaModel.model_class_reverse_map
+    d = Model.model_class_reverse_map
     if view_model_name in d:
         return d[view_model_name]
     else:
         raise KeyError("View model name '%s' not found" % view_model_name)
 
-class MetaModel(MetaHasProps):
-    ''' Specialize the construction of |Model| classes.
-
-    This class is a `metaclass`_ for |Model| that is responsible for
-    automatically cataloging all Bokeh models that get defined, so that the
-    serialization machinery between Bokeh and BokehJS can function properly.
-
-    .. note::
-        It is worth pointing out explicitly that this relies on the rules
-        for Metaclass inheritance in Python.
-
-    Bokeh works by replicating Python model objects (e.g. plots, ranges,
-    data sources, which are all |HasProps| subclasses) into BokehJS. In the
-    case of using a Bokeh server, the Bokeh model objects can also be
-    synchronized bidirectionally. This is accomplished by serializing the
-    models to and from a JSON format, that includes the name of the model type
-    as part of the payload, as well as a unique ID, and all the attributes:
-
-    .. code-block:: javascript
-
-        {
-            type: "Plot",
-            id: 100032,
-            attributes: { ... }
-        }
-
-    Typically the type name is inferred automatically from the Python class
-    name, and is set as the ``__view_model__`` class attribute on the Model
-    class that is create. But it is also possible to override this value
-    explicitly:
-
-    .. code-block:: python
-
-        class Foo(Model): pass
-
-        class Bar(Model):
-            __view_model__ == "Quux"
-
-    This metaclass will raise an error if two Bokeh models are created that
-    attempt to have the same view model name. The only exception made is if
-    one of the models has a custom ``__implementation__`` in its class
-    definition.
-
-    This metaclass also handles subtype relationships between Bokeh models.
-    Occasionally it may be necessary for multiple class types on the Python
-    side to resolve to the same type on the BokehJS side. This is called
-    subtyping, and is expressed through a ``__subtype__`` class attribute on
-    a model:
-
-    .. code-block:: python
-
-        class Foo(Model): pass
-
-        class Bar(Foo):
-            __view_model__ = "Foo"
-            __subtype__ = "Bar"
-
-    In this case, python instances of ``Foo`` and ``Bar`` will both resolve to
-    ``Foo`` models in BokehJS. In the context of a Bokeh server application,
-    the original python types will be faithfully round-tripped. (Without the
-    ``__subtype__`` specified, the above code would raise an error due to
-    duplicate view model names.)
-
-    .. _metaclass: https://docs.python.org/3/reference/datamodel.html#metaclasses
-
-    '''
-
-    model_class_reverse_map = {}
-
-    def __new__(meta_cls, class_name, bases, class_dict):
-        '''
-
-        Raises:
-            Warning
-
-        '''
-
-        # use an explicitly provided view model name if there is one
-        if "__view_model__" not in class_dict:
-            class_dict["__view_model__"] = class_name
-
-        # call the parent metaclass to create the new model type
-        newcls = super(MetaModel, meta_cls).__new__(meta_cls, class_name, bases, class_dict)
-
-        # update the mapping of view model names to classes, checking for any duplicates
-        # and handling any subtype relationships or custom implementations
-        entry = class_dict.get("__subtype__", class_dict["__view_model__"])
-        if entry in MetaModel.model_class_reverse_map and not hasattr(newcls, "__implementation__"):
-            raise Warning("Duplicate __view_model__ or __subtype__ declaration of '%s' for " \
-                          "class %s.  Previous definition: %s" % \
-                          (entry, class_name,
-                           MetaModel.model_class_reverse_map[entry]))
-        MetaModel.model_class_reverse_map[entry] = newcls
-
-        return newcls
+#-----------------------------------------------------------------------------
+# Dev API
+#-----------------------------------------------------------------------------
 
 _HTML_REPR = """
 <script>
@@ -208,22 +168,80 @@ _HTML_REPR = """
 </script>
 """
 
-class Model(with_metaclass(MetaModel, HasProps, PropertyCallbackManager, EventCallbackManager)):
+@abstract
+class Model(HasProps, PropertyCallbackManager, EventCallbackManager):
     ''' Base class for all objects stored in Bokeh  |Document| instances.
 
     '''
 
+    model_class_reverse_map = {}
+
+    @classmethod
+    def __init_subclass__(cls):
+        super().__init_subclass__()
+
+        # use an explicitly provided view model name if there is one
+        if "__view_model__" not in cls.__dict__:
+            cls.__view_model__ = cls.__name__
+        if "__view_module__" not in cls.__dict__:
+            cls.__view_module__ = cls.__module__
+
+        module = cls.__view_module__
+        model = cls.__dict__.get("__subtype__", cls.__view_model__)
+        impl = cls.__dict__.get("__implementation__", None)
+
+        head = module.split(".")[0]
+        if head == "bokeh" or head == "__main__" or impl is not None:
+            qualified = model
+        else:
+            qualified = module + "." + model
+
+        cls.__qualified_model__ = qualified
+
+        # update the mapping of view model names to classes, checking for any duplicates
+        # and handling any subtype relationships or custom implementations
+        previous = cls.model_class_reverse_map.get(qualified, None)
+        if previous is not None and not hasattr(cls, "__implementation__"):
+            raise Warning(f"Duplicate qualified model declaration of '{qualified}'. Previous definition: {previous}")
+        cls.model_class_reverse_map[qualified] = cls
+
+    def __new__(cls, *args, **kwargs):
+        obj =  super().__new__(cls)
+        obj._id = kwargs.pop("id", make_id())
+        obj._document = None
+        obj._temp_document = None
+        return obj
+
     def __init__(self, **kwargs):
-        self._id = kwargs.pop("id", make_id())
-        self._document = None
-        super(Model, self).__init__(**kwargs)
+
+        # "id" is popped from **kw in __new__, so in an ideal world I don't
+        # think it should be here too. But Python does this, so it is:
+        #
+        # class Foo(object):
+        #     def __new__(cls, *args, **kw):
+        #         obj = super().__new__(cls)
+        #         obj.bar = kw.pop("bar", 111)
+        #         print("__new__  :", id(kw), kw)
+        #         return obj
+        #     def __init__(self, **kw):
+        #         print("__init__ :", id(kw), kw)
+        #
+        # >>> f = Foo(bar=10)
+        # __new__  : 4405522296 {}
+        # __init__ : 4405522296 {'bar': 10}
+        kwargs.pop("id", None)
+
+        super().__init__(**kwargs)
         default_theme.apply_to_model(self)
 
     def __str__(self):
-        return "%s(id=%r, ...)" % (self.__class__.__name__, getattr(self, "_id", None))
+        return "%s(id=%r, ...)" % (self.__class__.__name__, getattr(self, "id", None))
 
     __repr__ = __str__
 
+    @property
+    def id(self):
+        return self._id
 
     name = String(help="""
     An arbitrary, user-supplied name for this model.
@@ -259,7 +277,7 @@ class Model(with_metaclass(MetaModel, HasProps, PropertyCallbackManager, EventCa
         [GlyphRenderer(id='1de4c3df-a83d-480a-899b-fb263d3d5dd9', ...)]
 
     Or simply a convenient way to attach any necessary metadata to a model
-    that can be accessed by CustomJS callbacks, etc.
+    that can be accessed by ``CustomJS`` callbacks, etc.
 
     .. note::
         No uniqueness guarantees or other conditions are enforced on any tags
@@ -270,14 +288,14 @@ class Model(with_metaclass(MetaModel, HasProps, PropertyCallbackManager, EventCa
 
     js_event_callbacks = Dict(String, List(Instance("bokeh.models.callbacks.CustomJS")),
     help="""
-    A mapping of event names to lists of CustomJS callbacks.
+    A mapping of event names to lists of ``CustomJS`` callbacks.
 
     Typically, rather then modifying this property directly, callbacks should be
     added using the ``Model.js_on_event`` method:
 
     .. code:: python
 
-        callback = CustomJS(code="console.log('tap event occured')")
+        callback = CustomJS(code="console.log('tap event occurred')")
         plot.js_on_event('tap', callback)
     """)
 
@@ -288,7 +306,7 @@ class Model(with_metaclass(MetaModel, HasProps, PropertyCallbackManager, EventCa
     """)
 
     js_property_callbacks = Dict(String, List(Instance("bokeh.models.callbacks.CustomJS")), help="""
-    A mapping of attribute names to lists of CustomJS callbacks, to be set up on
+    A mapping of attribute names to lists of ``CustomJS`` callbacks, to be set up on
     BokehJS side when the document is created.
 
     Typically, rather then modifying this property directly, callbacks should be
@@ -301,18 +319,24 @@ class Model(with_metaclass(MetaModel, HasProps, PropertyCallbackManager, EventCa
 
     """)
 
+    # Properties --------------------------------------------------------------
 
     @property
     def document(self):
         ''' The |Document| this model is attached to (can be ``None``)
 
         '''
+        if self._temp_document is not None:
+            return self._temp_document
         return self._document
 
     @property
     def ref(self):
-        ''' A Bokeh protocol "reference" to this model, i.e. a dict of the
-        form:
+        return dict(id=self._id)
+
+    @property
+    def struct(self):
+        ''' A Bokeh protocol "structure" of this model, i.e. a dict of the form:
 
         .. code-block:: python
 
@@ -324,17 +348,21 @@ class Model(with_metaclass(MetaModel, HasProps, PropertyCallbackManager, EventCa
         Additionally there may be a `subtype` field if this model is a subtype.
 
         '''
+        this = {
+            'type': self.__qualified_model__,
+            'id'  : self.id,
+        }
+
         if "__subtype__" in self.__class__.__dict__:
-            return {
-                'type'    : self.__view_model__,
-                'subtype' : self.__subtype__,
-                'id'      : self._id,
-            }
-        else:
-            return {
-                'type' : self.__view_model__,
-                'id'   : self._id,
-            }
+            # XXX: remove __subtype__ and this garbage at 2.0
+            parts = this["type"].split(".")
+            parts[-1] = self.__view_model__
+            this["type"] = ".".join(parts)
+            this["subtype"] = self.__subtype__
+
+        return this
+
+    # Public methods ----------------------------------------------------------
 
     def js_on_event(self, event, *callbacks):
 
@@ -350,8 +378,87 @@ class Model(with_metaclass(MetaModel, HasProps, PropertyCallbackManager, EventCa
             self.js_event_callbacks[event].append(callback)
 
 
+    def js_link(self, attr, other, other_attr, attr_selector=None):
+        ''' Link two Bokeh model properties using JavaScript.
+
+        This is a convenience method that simplifies adding a CustomJS callback
+        to update one Bokeh model property whenever another changes value.
+
+        Args:
+
+            attr (str) :
+                The name of a Bokeh property on this model
+
+            other (Model):
+                A Bokeh model to link to self.attr
+
+            other_attr (str) :
+                The property on ``other`` to link together
+
+            attr_selector (Union[int, str]) :
+                The index to link an item in a subscriptable ``attr``
+
+        Added in version 1.1
+
+        Raises:
+
+            ValueError
+
+        Examples:
+
+            This code with ``js_link``:
+
+            .. code :: python
+
+                select.js_link('value', plot, 'sizing_mode')
+
+            is equivalent to the following:
+
+            .. code:: python
+
+                from bokeh.models import CustomJS
+                select.js_on_change('value',
+                    CustomJS(args=dict(other=plot),
+                             code="other.sizing_mode = this.value"
+                    )
+                )
+
+            Additionally, to use attr_selector to attach the left side of a range slider to a plot's x_range:
+
+            .. code :: python
+
+                range_slider.js_link('value', plot.x_range, 'start', attr_selector=0)
+
+            which is equivalent to:
+
+            .. code :: python
+
+                from bokeh.models import CustomJS
+                range_slider.js_on_change('value',
+                    CustomJS(args=dict(other=plot.x_range),
+                             code="other.start = this.value[0]"
+                    )
+                )
+
+        '''
+        if attr not in self.properties():
+            raise ValueError("%r is not a property of self (%r)" % (attr, self))
+
+        if not isinstance(other, Model):
+            raise ValueError("'other' is not a Bokeh model: %r" % other)
+
+        if other_attr not in other.properties():
+            raise ValueError("%r is not a property of other (%r)" % (other_attr, other))
+
+        from bokeh.models import CustomJS
+
+        selector = f"[{attr_selector!r}]" if attr_selector is not None else ""
+        cb = CustomJS(args=dict(other=other), code=f"other.{other_attr} = this.{attr}{selector}")
+
+        self.js_on_change(attr, cb)
+
     def js_on_change(self, event, *callbacks):
-        ''' Attach a CustomJS callback to an arbitrary BokehJS model event.
+        ''' Attach a ``CustomJS`` callback to an arbitrary BokehJS model event.
 
         On the BokehJS side, change events for model properties have the
         form ``"change:property_name"``. As a convenience, if the event name
@@ -378,19 +485,21 @@ class Model(with_metaclass(MetaModel, HasProps, PropertyCallbackManager, EventCa
             raise ValueError("js_on_change takes an event name and one or more callbacks, got only one parameter")
 
         # handle any CustomJS callbacks here
-        from bokeh.models.callbacks import CustomJS
+        from bokeh.models import CustomJS
         if not all(isinstance(x, CustomJS) for x in callbacks):
             raise ValueError("not all callback values are CustomJS instances")
 
         if event in self.properties():
             event = "change:%s" % event
 
+        old = {k: [cb for cb in cbs] for k, cbs in self.js_property_callbacks.items()}
         if event not in self.js_property_callbacks:
             self.js_property_callbacks[event] = []
         for callback in callbacks:
             if callback in self.js_property_callbacks[event]:
                 continue
             self.js_property_callbacks[event].append(callback)
+        self.trigger('js_property_callbacks', old, self.js_property_callbacks)
 
     def layout(self, side, plot):
         '''
@@ -398,7 +507,7 @@ class Model(with_metaclass(MetaModel, HasProps, PropertyCallbackManager, EventCa
         '''
         try:
             return self in getattr(plot, side)
-        except:
+        except Exception:
             return []
 
     def on_change(self, attr, *callbacks):
@@ -406,15 +515,21 @@ class Model(with_metaclass(MetaModel, HasProps, PropertyCallbackManager, EventCa
 
         Args:
             attr (str) : an attribute name on this object
-            callback (callable) : a callback function to register
+            *callbacks (callable) : callback functions to register
 
         Returns:
             None
 
+        Example:
+
+        .. code-block:: python
+
+            widget.on_change('value', callback1, callback2, ..., callback_n)
+
         '''
         if attr not in self.properties():
             raise ValueError("attempted to add a callback on nonexistent %s.%s property" % (self.__class__.__name__, attr))
-        super(Model, self).on_change(attr, *callbacks)
+        super().on_change(attr, *callbacks)
 
     def references(self):
         ''' Returns all ``Models`` that this object has references to.
@@ -433,6 +548,7 @@ class Model(with_metaclass(MetaModel, HasProps, PropertyCallbackManager, EventCa
             seq[Model]
 
         '''
+        from .core.query import find
         return find(self.references(), selector)
 
     def select_one(self, selector):
@@ -464,6 +580,8 @@ class Model(with_metaclass(MetaModel, HasProps, PropertyCallbackManager, EventCa
             None
 
         '''
+        if isclass(selector) and issubclass(selector, Model):
+            selector = dict(type=selector)
         for obj in self.select(selector):
             for key, val in updates.items():
                 setattr(obj, key, val)
@@ -478,7 +596,7 @@ class Model(with_metaclass(MetaModel, HasProps, PropertyCallbackManager, EventCa
         need to separately have the full attributes of those
         other objects.
 
-        There's no corresponding from_json() because to
+        There's no corresponding ``from_json()`` because to
         deserialize an object is normally done in the context of a
         Document (since the Document can resolve references).
 
@@ -500,7 +618,7 @@ class Model(with_metaclass(MetaModel, HasProps, PropertyCallbackManager, EventCa
         will need to separately have the full attributes of those
         other objects.
 
-        There's no corresponding from_json_string() because to
+        There's no corresponding ``from_json_string()`` because to
         deserialize an object is normally done in the context of a
         Document (since the Document can resolve references).
 
@@ -513,7 +631,7 @@ class Model(with_metaclass(MetaModel, HasProps, PropertyCallbackManager, EventCa
 
         '''
         json_like = self._to_json_like(include_defaults=include_defaults)
-        json_like['id'] = self._id
+        json_like['id'] = self.id
         # serialize_json "fixes" the JSON from _to_json_like by converting
         # all types into plain JSON types # (it converts Model into refs,
         # for example).
@@ -527,7 +645,7 @@ class Model(with_metaclass(MetaModel, HasProps, PropertyCallbackManager, EventCa
         # The explicit assumption here is that hinted events do not need to
         # go through all the same invalidation steps. Currently this is the
         # case for ColumnsStreamedEvent and ColumnsPatchedEvent. However,
-        # this may need to be further refined in the future, if the a
+        # this may need to be further refined in the future, if the
         # assumption does not hold for future hinted events (e.g. the hint
         # could specify explicitly whether to do normal invalidation or not)
         if hint is None:
@@ -540,7 +658,7 @@ class Model(with_metaclass(MetaModel, HasProps, PropertyCallbackManager, EventCa
                 if dirty['count'] > 0:
                     self._document._invalidate_all_models()
         # chain up to invoke callbacks
-        super(Model, self).trigger(attr, old, new, hint=hint, setter=setter)
+        super().trigger(attr, old, new, hint=hint, setter=setter)
 
     def _attach_document(self, doc):
         ''' Attach a model to a Bokeh |Document|.
@@ -555,6 +673,13 @@ class Model(with_metaclass(MetaModel, HasProps, PropertyCallbackManager, EventCa
         self._document = doc
         self._update_event_callbacks()
 
+    @staticmethod
+    def _clear_extensions():
+        Model.model_class_reverse_map = {
+            k: v for k, v in Model.model_class_reverse_map.items()
+            if getattr(v, "__implementation__", None) is None
+        }
+
     def _detach_document(self):
         ''' Detach a model from a Bokeh |Document|.
 
@@ -575,7 +700,7 @@ class Model(with_metaclass(MetaModel, HasProps, PropertyCallbackManager, EventCa
         That's what "json like" means.
 
         This method should be considered "private" or "protected",
-        for use internal to Bokeh; use to_json() instead because
+        for use internal to Bokeh; use ``to_json()`` instead because
         it gives you only plain JSON-compatible types.
 
         Args:
@@ -649,32 +774,9 @@ class Model(with_metaclass(MetaModel, HasProps, PropertyCallbackManager, EventCa
 
         return html
 
-    def _repr_pretty(self, p, cycle):
-        '''
-
-        '''
-        name = "%s.%s" % (self.__class__.__module__, self.__class__.__name__)
-        _id = getattr(self, "_id", None)
-
-        if cycle:
-            p.text(name)
-            p.text('(id=')
-            p.pretty(_id)
-            p.text(', ...)')
-        else:
-            with p.group(4, '%s(' % name, ')'):
-                props = self.properties_with_values().items()
-                sorted_props = sorted(props, key=itemgetter(0))
-                all_props = [('id', _id)] + sorted_props
-                for i, (prop, value) in enumerate(all_props):
-                    if i == 0:
-                        p.breakable('')
-                    else:
-                        p.text(',')
-                        p.breakable()
-                    p.text(prop)
-                    p.text('=')
-                    p.pretty(value)
+#-----------------------------------------------------------------------------
+# Private API
+#-----------------------------------------------------------------------------
 
 def _visit_immediate_value_references(value, visitor):
     ''' Visit all references to another Model without recursing into any
@@ -692,7 +794,6 @@ def _visit_immediate_value_references(value, visitor):
 
 _common_types = {int, float, str}
 
-
 def _visit_value_and_its_immediate_references(obj, visitor):
     ''' Recurse down Models, HasProps, and Python containers
 
@@ -708,7 +809,7 @@ def _visit_value_and_its_immediate_references(obj, visitor):
         for item in obj:
             _visit_value_and_its_immediate_references(item, visitor)
     elif issubclass(typ, dict):
-        for key, value in iteritems(obj):
+        for key, value in obj.items():
             _visit_value_and_its_immediate_references(key, visitor)
             _visit_value_and_its_immediate_references(value, visitor)
     elif issubclass(typ, HasProps):
@@ -717,3 +818,8 @@ def _visit_value_and_its_immediate_references(obj, visitor):
         else:
             # this isn't a Model, so recurse into it
             _visit_immediate_value_references(obj, visitor)
+
+
+#-----------------------------------------------------------------------------
+# Code
+#-----------------------------------------------------------------------------
