@@ -1,5 +1,5 @@
 #-----------------------------------------------------------------------------
-# Copyright (c) 2012 - 2019, Anaconda, Inc., and Bokeh Contributors.
+# Copyright (c) 2012 - 2020, Anaconda, Inc., and Bokeh Contributors.
 # All rights reserved.
 #
 # The full license is in the file LICENSE.txt, distributed with this software.
@@ -8,9 +8,7 @@
 #-----------------------------------------------------------------------------
 # Boilerplate
 #-----------------------------------------------------------------------------
-from __future__ import absolute_import, division, print_function
-
-import logging
+import logging # isort:skip
 log = logging.getLogger(__name__)
 
 #-----------------------------------------------------------------------------
@@ -20,17 +18,28 @@ log = logging.getLogger(__name__)
 # Standard library imports
 import warnings
 
-# External imports
-
 # Bokeh imports
 from ..core.has_props import abstract
-from ..core.properties import Any, Bool, ColumnData, Dict, Enum, Instance, Int, JSON, List, Seq, String
+from ..core.properties import (
+    JSON,
+    Any,
+    Bool,
+    ColumnData,
+    Dict,
+    Enum,
+    Instance,
+    Int,
+    List,
+    PandasDataFrame,
+    PandasGroupBy,
+    Seq,
+    String,
+)
 from ..model import Model
 from ..util.dependencies import import_optional
 from ..util.serialization import convert_datetime_array
 from ..util.warnings import BokehUserWarning
-
-from .callbacks import Callback, CustomJS
+from .callbacks import CustomJS
 from .filters import Filter
 from .selections import Selection, SelectionPolicy, UnionRenderers
 
@@ -41,14 +50,14 @@ pd = import_optional('pandas')
 #-----------------------------------------------------------------------------
 
 __all__ = (
-    'ServerSentDataSource',
     'AjaxDataSource',
     'CDSView',
     'ColumnarDataSource',
     'ColumnDataSource',
     'DataSource',
     'GeoJSONDataSource',
-    'RemoteSource',
+    'ServerSentDataSource',
+    'WebSource',
 )
 
 #-----------------------------------------------------------------------------
@@ -61,18 +70,10 @@ class DataSource(Model):
 
     '''
 
-    selected = Instance(Selection, default=lambda: Selection(), help="""
+    selected = Instance(Selection, default=lambda: Selection(), readonly=True, help="""
     A Selection that indicates selected indices on this ``DataSource``.
     """)
 
-    callback = Instance(Callback, help="""
-    A callback to run in the browser whenever the selection is changed.
-
-    .. note:
-        This property is left for backwards compatibility, but may be deprecated
-        in the future. Prefer ``source.selected.js_on_change(...)`` for new code.
-    """)
-
 @abstract
 class ColumnarDataSource(DataSource):
     ''' A base class for data source types, which can be mapped onto
@@ -156,9 +157,17 @@ class ColumnDataSource(ColumnarDataSource):
     '''
 
     data = ColumnData(String, Seq(Any), help="""
-    Mapping of column names to sequences of data. The data can be, e.g,
+    Mapping of column names to sequences of data. The columns can be, e.g,
     Python lists or tuples, NumPy arrays, etc.
-    """).asserts(lambda _, data: len(set(len(x) for x in data.values())) <= 1,
+
+    The .data attribute can also be set from Pandas DataFrames or GroupBy
+    objects. In these cases, the behaviour is identical to passing the objects
+    to the ``ColumnDataSource`` initializer.
+    """).accepts(
+        PandasDataFrame, lambda x: ColumnDataSource._data_from_df(x)
+    ).accepts(
+        PandasGroupBy, lambda x: ColumnDataSource._data_from_groupby(x)
+    ).asserts(lambda _, data: len(set(len(x) for x in data.values())) <= 1,
                  lambda obj, name, data: warnings.warn(
                     "ColumnDataSource's columns must be of the same length. " +
                     "Current lengths: %s" % ", ".join(sorted(str((k, len(v))) for k, v in data.items())), BokehUserWarning))
@@ -181,7 +190,7 @@ class ColumnDataSource(ColumnarDataSource):
                 raw_data = self._data_from_groupby(raw_data)
             else:
                 raise ValueError("expected a dict or pandas.DataFrame, got %s" % raw_data)
-        super(ColumnDataSource, self).__init__(**kw)
+        super().__init__(**kw)
         self.data.update(raw_data)
 
     @property
@@ -223,7 +232,7 @@ class ColumnDataSource(ColumnarDataSource):
             _df.index = pd.Index(_df.index.values, name=index_name)
         _df.reset_index(inplace=True)
 
-        tmp_data = {c: v.values for c, v in _df.iteritems()}
+        tmp_data = {c: v.values for c, v in _df.items()}
 
         new_data = {}
         for k, v in tmp_data.items():
@@ -359,7 +368,6 @@ class ColumnDataSource(ColumnarDataSource):
         try:
             del self.data[name]
         except (ValueError, KeyError):
-            import warnings
             warnings.warn("Unable to find column '%s' in data source" % name)
 
     def stream(self, new_data, rollover=None):
@@ -466,7 +474,7 @@ class ColumnDataSource(ColumnarDataSource):
             newkeys = set(_df.columns)
             index_name = ColumnDataSource._df_index_name(_df)
             newkeys.add(index_name)
-            new_data = dict(_df.iteritems())
+            new_data = dict(_df.items())
             new_data[index_name] = _df.index.values
         else:
             newkeys = set(new_data.keys())
@@ -734,27 +742,13 @@ class WebSource(ColumnDataSource):
     A URL to to fetch data from.
     """)
 
-@abstract
-class RemoteSource(WebSource):
-    ''' Base class for remote column data sources that can update from data
-    URLs at prescribed time intervals.
-
-    .. note::
-        This base class is typically not useful to instantiate on its own.
-
-    '''
-
-    polling_interval = Int(help="""
-    A polling interval (in milliseconds) for updating data source.
-    """)
-
 class ServerSentDataSource(WebSource):
     ''' A data source that can populate columns by receiving server sent
     events endpoints.
 
     '''
 
-class AjaxDataSource(RemoteSource):
+class AjaxDataSource(WebSource):
     ''' A data source that can populate columns by making Ajax calls to REST
     endpoints.
 
@@ -777,10 +771,18 @@ class AjaxDataSource(RemoteSource):
     callback can be provided to convert the REST response into Bokeh format,
     via the ``adapter`` property of this data source.
 
+    Initial data can be set by specifying the ``data`` property directly.
+    This is necessary when used in conjunction with a ``FactorRange``, even
+    if the columns in `data`` are empty.
+
     A full example can be seen at :bokeh-tree:`examples/howto/ajax_source.py`
 
     '''
 
+    polling_interval = Int(help="""
+    A polling interval (in milliseconds) for updating data source.
+    """)
+
     method = Enum('POST', 'GET', help="""
     Specify the HTTP method to use for the Ajax request (GET or POST)
     """)
